﻿using System.Diagnostics;
using BbqpCmcs.Types;

namespace BbqpCmcs.AtomicOperators
{
	class InternalBbqpSolution
	{
		private readonly bool[] x;
		private readonly bool[] y;
		private readonly long[] rowSums;
		private readonly long[] colSums;
		private long cSum;
		private long dSum;
		private long objective;

		private readonly BbqpInstance instance;

		public InternalBbqpSolution(bool[] x, bool[] y, long[] rowSums, long[] colSums, long cSum, long dSum, long objective, BbqpInstance instance)
		{
			this.x = x;
			this.y = y;
			this.rowSums = rowSums;
			this.colSums = colSums;
			this.cSum = cSum;
			this.dSum = dSum;
			this.objective = objective;
			this.instance = instance;

			Debug.Assert(x.Length == instance.M);
			Debug.Assert(y.Length == instance.N);
			Debug.Assert(rowSums.Length == instance.M);
			Debug.Assert(colSums.Length == instance.N);
			Debug.Assert(TestIntegrity());
		}

		public InternalBbqpSolution(BbqpInstance instance)
		{
			this.instance = instance;
			x = new bool[instance.M];
			y = new bool[instance.N];
			rowSums = new long[instance.M];
			colSums = new long[instance.N];
			instance.CopyC(rowSums);
			instance.CopyD(colSums);

			Debug.Assert(TestIntegrity());
		}

		public static InternalBbqpSolution FromBbqpSolution(BbqpSolution solution)
		{
			bool[] x, y;
			solution.CopyValues(out x, out y);
			return new InternalBbqpSolution(solution.Instance, x, y);
		}

		public InternalBbqpSolution(BbqpInstance instance, bool[] x, bool[] y)
		{
			this.instance = instance;
			this.x = x;
			this.y = y;

			rowSums = new long[instance.M];
			colSums = new long[instance.N];
			instance.CopyC(rowSums);
			instance.CopyD(colSums);

			for (int i = 0; i < instance.M; i++)
				if (x[i])
				{
					for (int j = 0; j < instance.N; j++)
						colSums[j] += instance[i, j];

					cSum += instance.GetC(i);
				}

			objective = cSum;

			for (int j = 0; j < instance.N; j++)
				if (y[j])
				{
					for (int i = 0; i < instance.M; i++)
						rowSums[i] += instance[i, j];

					objective += colSums[j];
					dSum += instance.GetD(j);
				}

			Debug.Assert(TestIntegrity());
		}

		public BbqpInstance Instance
		{
			get { return instance; }
		}

		public long Objective
		{
			get { return objective; }
		}

		public bool GetX(int i)
		{
			return x[i];
		}

		public bool GetY(int j)
		{
			return y[j];
		}

		public long CSum
		{
			get { return cSum; }
		}

		public long DSum
		{
			get { return dSum; }
		}

		public BbqpSolution GetSolution(TimeInterval runningTime)
		{
			return new BbqpSolution(instance, (bool[])x.Clone(), (bool[])y.Clone(), runningTime, false);
		}

		public void SetX(int i, bool newValue)
		{
			if (x[i] == newValue)
				return;

			x[i] = newValue;
			if (newValue)
			{
				objective += rowSums[i];
				for (int j = instance.N - 1; j >= 0; j--)
					colSums[j] += instance[i, j];

				cSum += instance.GetC(i);
			}
			else
			{
				objective -= rowSums[i];
				for (int j = instance.N - 1; j >= 0; j--)
					colSums[j] -= instance[i, j];

				cSum -= instance.GetC(i);
			}

			//Debug.Assert(TestIntegrity());
		}

		public void SetY(int j, bool newValue)
		{
			if (y[j] == newValue)
				return;

			y[j] = newValue;
			if (newValue)
			{
				objective += colSums[j];
				for (int i = instance.M - 1; i >= 0; i--)
					rowSums[i] += instance[i, j];

				dSum += instance.GetD(j);
			}
			else
			{
				objective -= colSums[j];
				for (int i = instance.M - 1; i >= 0; i--)
					rowSums[i] -= instance[i, j];

				dSum -= instance.GetD(j);
			}

			//Debug.Assert(TestIntegrity());
		}

		private bool TestIntegrity()
		{
			long obj1 = 0;
			long obj2 = 0;
			long obj3 = 0;

			for (int i = 0; i < instance.M; i++)
				for (int j = 0; j < instance.N; j++)
					if (x[i] && y[j])
						obj3 += instance[i, j];

			long cSumCheck = 0;
			for (int i = 0; i < instance.M; i++)
				if (x[i])
					cSumCheck += instance.GetC(i);

			if (cSumCheck != cSum)
				return false;
					
			obj3 += cSumCheck;

			long dSumCheck = 0;
			for (int j = 0; j < instance.N; j++)
				if (y[j])
					dSumCheck += instance.GetD(j);

			if (dSumCheck != dSum)
				return false;

			obj3 += dSumCheck;

			for (int i = 0; i < instance.M; i++)
			{
				long sum = instance.GetC(i);
				for (int j = 0; j < instance.N; j++)
					if (y[j])
						sum += instance[i, j];

				if (sum != rowSums[i])
					return false;

				if (x[i])
					obj1 += rowSums[i];
			}

			for (int j = 0; j < instance.N; j++)
				if (y[j])
					obj1 += instance.GetD(j);

			for (int j = 0; j < instance.N; j++)
			{
				long sum = instance.GetD(j);
				for (int i = 0; i < instance.M; i++)
					if (x[i])
						sum += instance[i, j];

				if (sum != colSums[j])
					return false;

				if (y[j])
					obj2 += colSums[j];
			}

			for (int i = 0; i < instance.M; i++)
				if (x[i])
					obj2 += instance.GetC(i);

			if (obj3 != objective)
				return false;

			if (obj1 != objective)
				return false;

			if (obj2 != objective)
				return false;

			return true;
		}

		public long GetRowSum(int i)
		{
			return rowSums[i];
		}

		public long GetColSum(int j)
		{
			return colSums[j];
		}

		public bool[] CopyX()
		{
			return (bool[])x.Clone();
		}

		public void CopyX(bool[] destinationX)
		{
			x.CopyTo(destinationX, 0);
		}

		public bool[] CopyY()
		{
			return (bool[])y.Clone();
		}

		public void CopyY(bool[] destinationY)
		{
			y.CopyTo(destinationY, 0);
		}

		public long[] CopyColSums()
		{
			return (long[])colSums.Clone();
		}

		public long[] CopyRowSums()
		{
			return (long[])rowSums.Clone();
		}
	}
}
