﻿using System;
using System.Diagnostics;

namespace BbqpCmcs.AtomicOperators
{
	class WalkSat : AtomicOperator
	{
		private readonly int tournamentSize;
		private readonly int numberOfIterations;

		public WalkSat(int tournamentSize, int numberOfIterations)
		{
			this.tournamentSize = tournamentSize;
			this.numberOfIterations = numberOfIterations;
			Name = "WS" + tournamentSize;
		}

/*
		struct CellToBeFixed
		{
			private int row;
			private int col;
			private int gain;
		}
*/

		public override void Run(InternalBbqpSolution solution, Random random)
		{
			var instance = solution.Instance;

			var x = solution.CopyX();
			var y = solution.CopyY();

			long[] positivesRow = new long[instance.M];
			long[] negativeRow = new long[instance.M];
			long[] positivesCol = new long[instance.N];
			long[] negativeCol = new long[instance.N];

			for (int j = 0; j < instance.N; j++)
			{
				var d = instance.GetD(j);
				if (d > 0)
					positivesCol[j] += d;
				else
					negativeCol[j] -= d;
			}

			for (int i = 0; i < instance.M; i++)
			{
				if (!x[i])
					continue;

				for (int j = 0; j < instance.N; j++)
				{
					var q = instance[i, j];
					if (q > 0)
						positivesCol[j] += q;
					else
						negativeCol[j] -= q;
				}
			}
			
			for (int i = 0; i < instance.M; i++)
			{
				var c = instance.GetC(i);
				if (c > 0)
					positivesRow[i] += c;
				else
					negativeRow[i] -= c;
			}

			for (int j = 0; j < instance.N; j++)
			{
				if (!y[j])
					continue;

				for (int i = 0; i < instance.M; i++)
				{
					var q = instance[i, j];
					if (q > 0)
						positivesRow[i] += q;
					else
						negativeRow[i] -= q;
				}
			}


			int collectedForTournament = 0;
			int bestRow = -1;
			int bestCol = -1;
			long bestGain = -1;

			for (int iteration = 0; iteration < numberOfIterations; iteration++)
			{
				int row = random.Next(instance.M);
				int col = random.Next(instance.N);
				var q = instance[row, col];

				if (q > 0)
				{
					if (!x[row] || !y[col])
					{
						long gain;
						if (!x[row])
						{
							gain = positivesRow[row] - q;
							if (!y[row])
								gain += positivesCol[row];
						}
						else
							gain = positivesCol[col];

						gain += q * 0;

						if (gain > bestGain)
						{
							bestGain = gain;
							bestRow = !x[row] ? row : -1;
							bestCol = !y[col] ? col : -1;
						}

						collectedForTournament++;
					}
				}
				else if (q < 0)
				{
					if (x[row] && y[col])
					{
						{ // disable x[row]
							long gain = negativeRow[row];
							gain -= q * 0;

							if (gain > bestGain)
							{
								bestGain = gain;
								bestRow = row;
								bestCol = -1;
							}
						}

						{ // disable y[col]
							long gain = negativeCol[col];
							gain -= q * 0;
							if (gain > bestGain)
							{
								bestGain = gain;
								bestRow = -1;
								bestCol = col;
							}
						}

						collectedForTournament++;
					}
				}


				if (collectedForTournament >= tournamentSize)
				{
					if (bestRow >= 0)
						x[bestRow] = !x[bestRow];

					if (bestCol >= 0)
						y[bestCol] = !y[bestCol];

					if (bestRow >= 0)
					{
						if (x[bestRow]) // enable row
						{
							for (int j = 0; j < instance.N; j++)
							{
								var q1 = instance[bestRow, j];
								if (q1 > 0)
									positivesCol[j] += q1;
								else
									negativeCol[j] -= q1;
							}
						}
						else
						{
							// disable row
							for (int j = 0; j < instance.N; j++)
							{
								var q1 = instance[bestRow, j];
								if (q1 > 0)
									positivesCol[j] -= q1;
								else
									negativeCol[j] += q1;
							}
						}
					}

					if (bestCol >= 0)
					{
						if (y[bestCol]) // enable col
						{
							for (int i = 0; i < instance.M; i++)
							{
								var q1 = instance[i, bestCol];
								if (q1 > 0)
									positivesRow[i] += q1;
								else
									negativeRow[i] -= q1;
							}
						}
						else
						{
							// disable col
							for (int i = 0; i < instance.M; i++)
							{
								var q1 = instance[i, bestCol];
								if (q1 > 0)
									positivesRow[i] -= q1;
								else
									negativeRow[i] += q1;
							}
						}
					}

#if DEBUG
					for (int i = 0; i < instance.M; i++)
					{
						long positive = Math.Max(0, instance.GetC(i));
						long negative = Math.Max(0, -instance.GetC(i));
						for (int j = 0; j < instance.N; j++)
						{
							if (y[j])
							{
								if (instance[i, j] > 0)
									positive += instance[i, j];
								else
									negative -= instance[i, j];
							}
						}

						Debug.Assert(positive == positivesRow[i]);
						Debug.Assert(negative == negativeRow[i]);
					}

					for (int j = 0; j < instance.N; j++)
					{
						long positive = Math.Max(0, instance.GetD(j));
						long negative = Math.Max(0, -instance.GetD(j));
						for (int i = 0; i < instance.M; i++)
						{
							if (x[i])
							{
								if (instance[i, j] > 0)
									positive += instance[i, j];
								else
									negative -= instance[i, j];
							}
						}

						Debug.Assert(positive == positivesCol[j]);
						Debug.Assert(negative == negativeCol[j]);
					}
#endif

					collectedForTournament = 0;
					bestGain = -1;
					bestRow = -1; // Not necessary; just in case
					bestCol = -1; // Not necessary; just in case
				}
			}

			for (int i = 0; i < instance.M; i++)
				solution.SetX(i, x[i]);

			for (int j = 0; j < instance.N; j++)
				solution.SetY(j, y[j]);
		}

		public override bool FindsLocalMinimum
		{
			get { return false; }
		}

		public override bool CanWorsen
		{
			get { return true; }
		}

		public override bool IsDeterministicImprovement
		{
			get { return false; }
		}
	}
}
