﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using BbqpCmcs.AtomicOperators;
using BbqpCmcs.Types;

namespace BbqpCmcs
{
	internal class ConditionalAtomicMarkovChainHh
	{
		private readonly AtomicOperator[] lowLevelHeuristics;
		private readonly double[][] weightsImproved;
		private readonly double[][] weightsUnimproved;
		private readonly TimeInterval timeLimit;
		private readonly int iterationsLimit;
		private readonly int[][] improvedCounters;
		private readonly int[][] unimprovedCounters;

		public ConditionalAtomicMarkovChainHh(
			AtomicOperator[] lowLevelHeuristics,
			double[][] weightsImproved,
			double[][] weightsUnimproved,
			int iterationsLimit)
			: this(lowLevelHeuristics, weightsImproved, weightsUnimproved, TimeInterval.Zero, iterationsLimit)
		{
		}

		public ConditionalAtomicMarkovChainHh(
			AtomicOperator[] lowLevelHeuristics,
			double[][] weightsImproved,
			double[][] weightsUnimproved,
			TimeInterval timeLimit,
			int iterationsLimit = 0,
			int[][] improvedCounters = null,
			int[][] unimprovedCounters = null)
		{
			this.lowLevelHeuristics = lowLevelHeuristics;
			this.weightsImproved = weightsImproved;
			this.weightsUnimproved = weightsUnimproved;
			this.timeLimit = timeLimit;
			this.iterationsLimit = iterationsLimit;
			this.improvedCounters = improvedCounters;
			this.unimprovedCounters = unimprovedCounters;

			Debug.Assert((iterationsLimit == 0) != timeLimit.IsZero);
			Debug.Assert(this.weightsImproved.Length == lowLevelHeuristics.Length);
			Debug.Assert(this.weightsImproved.All(w => w.Length == lowLevelHeuristics.Length));
			Debug.Assert(this.weightsUnimproved.Length == lowLevelHeuristics.Length);
			Debug.Assert(this.weightsUnimproved.All(w => w.Length == lowLevelHeuristics.Length));

			if (iterationsLimit == 0)
				iterationsLimit = int.MaxValue;
		}


		public static ConditionalAtomicMarkovChainHh FromFile(
			AtomicOperator[] lowLevelHeuristics,
			string filename,
			int iterationsLimit,
			bool enableCounters = false)
		{
			return FromFile(lowLevelHeuristics, filename, TimeInterval.Zero, iterationsLimit, enableCounters);
		}

		public static ConditionalAtomicMarkovChainHh FromFilePreservingOperators(AtomicOperator[] lowLevelHeuristics, string filename, TimeInterval timeLimit, int iterationsLimit = 0, bool enableCounters = false)
		{
			using (var reader = File.OpenText(filename))
			{
				if (Helper.ReadLine(reader, true).Trim().ToLower() != "operators:")
					throw new LabException("The file is supposed to start with 'Operators:'.");

				var translation = new List<int>();
				while (true)
				{
					var heuristic = reader.ReadLine();
					if (string.IsNullOrWhiteSpace(heuristic))
						break;

					heuristic = heuristic.Trim().ToLower();

					var index = Array.FindIndex(lowLevelHeuristics, op => op.Name.ToLower() == heuristic);
					translation.Add(index);
				}

				if (Helper.ReadLine(reader, true).Trim().ToLower() != "improved:")
					throw new LabException("The first matrix header is supposed to be 'Improved:'.");

				var improved = ReadMatrix(lowLevelHeuristics.Length, reader, translation);

	        
				if (Helper.ReadLine(reader, true).Trim().ToLower() != "unimproved:")
					throw new LabException("The first matrix header is supposed to be 'Unimproved:'.");

				var unimproved = ReadMatrix(lowLevelHeuristics.Length, reader, translation);

				if (enableCounters)
				{
					var improvedCounters = new int[lowLevelHeuristics.Length][];
					var unimprovedCounters = new int[lowLevelHeuristics.Length][];
					for (int i = 0; i < lowLevelHeuristics.Length; i++)
					{
						improvedCounters[i] = new int[lowLevelHeuristics.Length];
						unimprovedCounters[i] = new int[lowLevelHeuristics.Length];
					}

					return new ConditionalAtomicMarkovChainHh(
						lowLevelHeuristics,
						improved,
						unimproved,
						timeLimit,
						iterationsLimit,
						improvedCounters,
						unimprovedCounters);
				}

				return new ConditionalAtomicMarkovChainHh(
					lowLevelHeuristics,
					improved,
					unimproved,
					timeLimit,
					iterationsLimit);
			}
		}

	    public AtomicOperator[] Operators
	    {
	        get { return lowLevelHeuristics; }
	    }

	    public double[][] WeightsImproved
	    {
	        get { return weightsImproved; }
	    }

	    public double[][] WeightsUnimproved
	    {
	        get { return weightsUnimproved; }
	    }

	    public static ConditionalAtomicMarkovChainHh FromFile(AtomicOperator[] allOperators, string filename, TimeInterval timeLimit, int iterationsLimit = 0, bool enableCounters = false)
		{
			using (var reader = File.OpenText(filename))
			{
				if (Helper.ReadLine(reader, true).Trim().ToLower() != "operators:")
					throw new LabException("The file is supposed to start with 'Operators:'.");

				var operators = new List<AtomicOperator>();
				while (true)
				{
					var heuristic = reader.ReadLine();
					if (string.IsNullOrWhiteSpace(heuristic))
						break;

					heuristic = heuristic.Trim().ToLower();

					var h = Array.Find(allOperators, op => op.Name.ToLower() == heuristic);
					if (h == null)
						throw new LabException("Can't recognise operator {0}.", heuristic);
					operators.Add(h);
				}

				if (Helper.ReadLine(reader, true).Trim().ToLower() != "improved:")
					throw new LabException("The first matrix header is supposed to be 'Improved:'.");

				var improved = ReadMatrix(operators.Count, reader);

	        
				if (Helper.ReadLine(reader, true).Trim().ToLower() != "unimproved:")
					throw new LabException("The first matrix header is supposed to be 'Unimproved:'.");

				var unimproved = ReadMatrix(operators.Count, reader);

				if (enableCounters)
				{
					var improvedCounters = new int[operators.Count][];
					var unimprovedCounters = new int[operators.Count][];
					for (int i = 0; i < operators.Count; i++)
					{
						improvedCounters[i] = new int[operators.Count];
						unimprovedCounters[i] = new int[operators.Count];
					}

					return new ConditionalAtomicMarkovChainHh(
						operators.ToArray(),
						improved,
						unimproved,
						timeLimit,
						iterationsLimit,
						improvedCounters,
						unimprovedCounters);
				}

				return new ConditionalAtomicMarkovChainHh(
					operators.ToArray(),
					improved,
					unimproved,
					timeLimit,
					iterationsLimit);
			}
		}

		private static double[][] ReadMatrix(int k, TextReader reader, List<int> translation)
		{
			double[][] result = new double[k][];
			for (int i = 0; i < k; i++)
				result[i] = new double[k];

			for (int i = 0; i < translation.Count; i++)
			{
				var line = Helper.ReadLine(reader, true);
				var strings = line.Split(new[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
				if (strings.Length != translation.Count)
					throw new LabException("{0} elements in line expected but {1} found.", translation.Count, strings.Length);

				if (translation[i] < 0)
					continue;

				for (int j = 0; j < translation.Count; j++)
					if (translation[j] >= 0)
						result[translation[i]][translation[j]] = double.Parse(strings[j]);
			}

			return result;
		}

		private static double[][] ReadMatrix(int k, TextReader reader)
		{
			double[][] result = new double[k][];
			for (int i = 0; i < k; i++)
				result[i] = new double[k];

			for (int i = 0; i < k; i++)
			{
				var line = Helper.ReadLine(reader, true);
				var strings = line.Split(new[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
				if (strings.Length != k)
					throw new LabException("{0} elements in line expected but {1} found.", k, strings.Length);

				for (int j = 0; j < k; j++)
					result[i][j] = double.Parse(strings[j]);
			}

			return result;
		}

		public void SaveToFile(string filename)
		{
			using (var writer = File.CreateText(filename))
			{
				writer.WriteLine("Operators:");
				foreach (var op in lowLevelHeuristics)
					writer.WriteLine(op.Name);

				writer.WriteLine();

				writer.WriteLine("Improved:");
				WriteMatrix(writer, weightsImproved);
				writer.WriteLine();

				writer.WriteLine("Unimproved:");
				WriteMatrix(writer, weightsUnimproved);
			}
		}

		private static void WriteMatrix(StreamWriter writer, double[][] weights)
		{
			foreach (var row in weights)
				Helper.WriteTabSeparatedLine(writer, row);
		}

		public long Run(InternalBbqpSolution solution, Random random, out bool[] x, out bool[] y)
		{
			var counter = AccurateRealTimeCounter.StartNew();
			int currentHeuristic = 0;
			//List<int> chain = new List<int>();
			long bestObjective = solution.Objective;
			long initialObjective = solution.Objective;
			x = null;
			y = null;

			int iterations = 0;
			while (true)
			{
				if (iterationsLimit == 0)
				{
					if (counter.CurrentElapsed >= timeLimit)
						break;
				}
				else
				{
					if (iterations++ > iterationsLimit)
						break;
				}
				//chain.Add(currentHeuristic);

				var oldObjective = solution.Objective;
				lowLevelHeuristics[currentHeuristic].Run(solution, random);
				bool improved = oldObjective < solution.Objective;

				if (solution.Objective > bestObjective)
				{
					bestObjective = solution.Objective;
					if (x == null)
					{
						x = solution.CopyX();
						y = solution.CopyY();
					}
					else
					{
						solution.CopyX(x);
						solution.CopyY(y);
					}
				}

				var weights = improved ? weightsImproved : weightsUnimproved;
				var newHeuristic = random.Roulette(weights[currentHeuristic]);

				var counters = improved ? improvedCounters : unimprovedCounters;
				if (counters != null)
					counters[currentHeuristic][newHeuristic]++;

				currentHeuristic = newHeuristic;
			}

			//if (Helper.EnsureDebugInfoInterval("hh-chain", TimeInterval.Second * 60))
			    //Helper.DebugMessage("{0}: {1}", chain.Count, chain.ListToString(" -> "));

			if (initialObjective < bestObjective && x == null)
				throw new LabException();

			return bestObjective;
		}

		public static AtomicOperator[] GetMoves(int repairFlaws, int repairMaxIterations, int mutationXSmallStrength, int mutationYSmallStrength, int mutationXLargeStrength, int mutationYLargeStrength)
		{
			return new AtomicOperator[]
			{
				new OptimiseX(),
				new OptimiseY(),
				new FlipX(),
				new FlipY(),
				new Repair(repairFlaws, repairMaxIterations),
				new MutationX(mutationXSmallStrength),
				new MutationY(mutationYSmallStrength),
				new MutationX(mutationXLargeStrength),
				new MutationY(mutationYLargeStrength)
			};
		}
	}
}