﻿using System;
using System.IO;
using BbqpCmcs.AtomicOperators;
using BbqpCmcs.Types;

namespace BbqpCmcs
{
	class Program
	{
		const string ConfigurationFilename = "2-row.cmc";

		static void Main(string[] args)
		{
			if (args.Length != 2)
			{
				Console.WriteLine("CMCS solver for BBQP");
				Console.WriteLine("Free for non-commercial use");
				Console.WriteLine("Description: https://arxiv.org/abs/1605.02038");
				Console.WriteLine("Contact: daniel.karapetyan@gmail.com");
				Console.WriteLine("Usage: BbqpCmcs.exe <instance file name> <time limit in seconds (real number)>");
				return;
			}

			var operators = ConditionalAtomicMarkovChainHh.GetMoves(
				repairFlaws: 10,
				repairMaxIterations: 100,
				mutationXSmallStrength: 4,
				mutationYSmallStrength: 4,
				mutationXLargeStrength: 16,
				mutationYLargeStrength: 16);

			var instanceName = args[0];
			var timeLimitSeconds = double.Parse(args[1]);

			var instance = BbqpInstance.GetInstance(instanceName, instanceName, null);

			Console.WriteLine(
				"Loaded {0} successfully.  Solving with time limit {1} sec.  Using configuration from file '{2}'",
				instanceName,
				timeLimitSeconds,
				ConfigurationFilename);

			var objective = Evaluate(new Random(), instance, operators, TimeInterval.Second*timeLimitSeconds);

			Console.WriteLine("Obtained objective {0}; solution saved to 'solution.sol'", objective);
		}



		private static long Evaluate(Random random, BbqpInstance instance, AtomicOperator[] operators, TimeInterval timeLimit)
		{
			var hh = ConditionalAtomicMarkovChainHh.FromFilePreservingOperators(operators, ConfigurationFilename, timeLimit);

			bool[] x = new bool[instance.M];
			bool[] y = new bool[instance.N];

			for (int i = 0; i < x.Length; i++)
				x[i] = random.NextBoolean();

			for (int j = 0; j < y.Length; j++)
				y[j] = random.NextBoolean();

			var solution = new InternalBbqpSolution(instance, x, y);

			long objective = solution.Objective;
			try
			{
				objective = hh.Run(solution, random, out x, out y);
				solution.GetSolution(timeLimit).SaveToFileForTheLibrary("solution.sol");
			}
			catch (Exception ex)
			{
				Helper.ErrorMessage("Exception while optimising: {0}", ex);
			}

			return objective;
		}

	}
}
