﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;

namespace BbqpCmcs.Types
{
	internal class BbqpInstance : Instance
	{
		private readonly int[][] q;
		private readonly long[] c, d;

		/*
				public BbqpInstance(string shortName, int n, int m, Action<int[][]> produce)
					: this(shortName, Produce(n, m, produce))
				{
				}
		*/

		private static int ReadInt(TextReader reader)
		{
			var line = reader.ReadLine();
			if (line == null)
				throw new LabException("End of stream reached.");

			try
			{
				return int.Parse(line);
			}
			catch (Exception)
			{
				throw new LabException("Can't parse {0}.", line);
			}
		}

		private static T[] ReadArray<T>(TextReader reader, int expectedCount, Func<string, T> parse)
		{
			var line = reader.ReadLine();
			if (line == null)
				throw new LabException("End of stream reached.");

			var strings = line.Split(new[] { ' ', '\t' }, StringSplitOptions.RemoveEmptyEntries);
			if (strings.Length != expectedCount)
				throw new LabException("{0} elements expected but {1} elements found.",
				                       expectedCount,
				                       strings.Length);

			var result = new T[expectedCount];
			try
			{
				for (int i = 0; i < strings.Length; i++)
					result[i] = parse(strings[i]);
			}
			catch (Exception)
			{
				throw new LabException("Can't parse a value.");
			}

			return result;
		}

		private static void WriteArray<T>(TextWriter writer, T[] array)
		{
			if (array.Length > 0)
				writer.Write(array[0]);

			for (int i = 1; i < array.Length; i++)
			{
				writer.Write(' ');
				writer.Write(array[i]);
			}

			writer.WriteLine();
		}

		private static BbqpInstance LoadFromFile(string uniqueName, string shortName, string filename)
		{
			try
			{
				var timeCounter = AccurateRealTimeCounter.StartNew();
				using (var reader = File.OpenText(filename))
				{
					int m = ReadInt(reader);
					int n = ReadInt(reader);

					var q = new int[m][];
					var c = ReadArray(reader, m, long.Parse);
					var d = ReadArray(reader, n, long.Parse);
					for (int i = 0; i < m; i++)
						q[i] = ReadArray(reader, n, int.Parse);

					timeCounter.Stop();
					//if (timeCounter.Elapsed > TimeInterval.Second)
						//Helper.InfoMessage("{0:0.0} sec spent on reading {1}.", timeCounter.Elapsed.Seconds, shortName);

					return new BbqpInstance(uniqueName, shortName, q, c, d);
				}
			}
			catch (Exception ex)
			{
				throw new LabException("Failed to read {0}.  The following error occured: {1}",
				                       filename,
				                       ex.Message);
			}
		}

		public void SaveToFile(string filename)
		{
			if (!BbqpHelper.GenerationOfInstancesAllowed)
				throw new LabException("Generation of instances is prohibited.  Check the BbqpHelper.GenerationOfInstancesAllowed constant.");

			try
			{
				using (var writer = File.CreateText(filename))
				{
					writer.WriteLine(M);
					writer.WriteLine(N);
					WriteArray(writer, c);
					WriteArray(writer, d);
					for (int i = 0; i < M; i++)
						WriteArray(writer, q[i]);
				}
			}
			catch (Exception ex)
			{
				throw new LabException("Failed to write {0}.  The following error occured: {1}",
				                       filename,
				                       ex.Message);
			}
		}

		public BbqpInstance ConvertTo01()
		{
			var newQ = new int[M][];
			for (int i = 0; i < M; i++)
			{
				newQ[i] = new int[N];
				for (int j = 0; j < N; j++)
					newQ[i][j] = this[i, j] * 4;
			}

			var newC = (long[])c.Clone();
			for (int i = 0; i < M; i++)
			{
				for (int j = 0; j < N; j++)
					newC[i] -= this[i, j];

				newC[i] *= 2;
			}

			var newD = (long[])d.Clone();
			for (int j = 0; j < N; j++)
			{
				for (int i = 0; i < M; i++)
					newD[j] -= this[i, j];

				newD[j] *= 2;
			}

			return new BbqpInstance(UniqueName, ShortName, newQ, newC, newD);
		}

		public BbqpInstance(string uniqueName, string shortName, int[][] q, long[] c, long[] d)
			: base(uniqueName, shortName)
		{
			this.q = q;
			this.c = c;
			this.d = d;

			//Debug.Assert(N >= M);
			Debug.Assert(c.Length >= M);
			Debug.Assert(d.Length >= N);
		}

		public BbqpInstance(string uniqueName, string shortName, int m, int n, Func<int, int, int> produceQ,
		                   Func<int, long> produceC, Func<int, long> produceD)
			: base(uniqueName, shortName)
		{
			q = new int[m][];
			for (int i = 0; i < m; i++)
			{
				q[i] = new int[n];
				for (int j = 0; j < n; j++)
					q[i][j] = produceQ(i, j);
			}

			c = new long[m];
			for (int i = 0; i < m; i++)
				c[i] = produceC(i);

			d = new long[n];
			for (int j = 0; j < n; j++)
				d[j] = produceD(j);
		}

		/*
				private static int[][] Produce(int n, int m, Action<int[][]> produce)
				{
					var q = new int[n][];
					for (int i = 0; i < n; i++)
						q[i] = new int[m];

					produce(q);

					return q;
				}
		*/

		/*
				public BbqpInstance(string shortName, int[][] q) : base(shortName, shortName)
				{
					this.q = q;
					Debug.Assert(q.Length > 0 && q[0].Length > 0);

					if (q.Length > q[0].Length)
						throw new LabException("The number of rows in Q exceeds the number of columns.");

		#if DEBUG
					int n = q[0].Length;
					for (int j = 1; j < q.Length; j++)
					{
						if (q[j].Length != n)
							throw new LabException("The matrix Q is not rectangular.");
					}
		#endif
				}
		*/

		public int this[int row, int col]
		{
			get { return q[row][col]; }
		}

		public int M
		{
			get { return q.Length; }
		}

		public int N
		{
			get { return q[0].Length; }
		}

		public long GetC(int i)
		{
			return c[i];
		}

		public long GetD(int j)
		{
			return d[j];
		}

		public void CopyC(long[] cCopy)
		{
			Debug.Assert(cCopy.Length == c.Length);
			c.CopyTo(cCopy, 0);
		}

		public long[] CopyC()
		{
			return (long[])c.Clone();
		}

		public void CopyD(long[] dCopy)
		{
			Debug.Assert(dCopy.Length == d.Length);
			d.CopyTo(dCopy, 0);
		}

		public long[] CopyD()
		{
			return (long[])d.Clone();
		}

		public string TypeName
		{
			get
			{
				var spaceIndex = 0;
				while (char.IsLetter(UniqueName[spaceIndex]))
				{
					spaceIndex++;
					if (spaceIndex >= UniqueName.Length)
						throw new LabException("Unsupported instance name: {0}.", UniqueName);
				}

				return UniqueName.Substring(0, spaceIndex);
			}
		}

		private static string GetFullFilename(string uniqueName)
		{
			return Helper.GetFullFilename(Helper.ProgramDirectory, uniqueName);
		}

		private static readonly Dictionary<string, BbqpInstance> Instances =
			new Dictionary<string, BbqpInstance>();

		public static BbqpInstance GetInstance(string uniqueName, string shortName, int m, int n,
		                                      Func<int, int, int> getQ, Func<int, long> getC,
		                                      Func<int, long> getD)
		{
			if (Instances.ContainsKey(uniqueName))
				return Instances[uniqueName];

			var filename = GetFullFilename(uniqueName);
			if (File.Exists(filename))
				return LoadFromFile(uniqueName, shortName, filename);

			if (!BbqpHelper.GenerationOfInstancesAllowed)
				throw new LabException("Instance '{0}' not found.  Generation of instances is prohibited.  Check the BbqpHelper.GenerationOfInstancesAllowed constant.", filename);

			var timeCounter = AccurateRealTimeCounter.StartNew();
			var instance = new BbqpInstance(uniqueName, shortName, m, n, getQ, getC, getD);
			timeCounter.Stop();
			//if (timeCounter.Elapsed > TimeInterval.Second)
				//Helper.DebugMessage("{0} spent on generating {1}.", timeCounter.Elapsed, shortName);
			//instance.SaveToFile(filename);

			Instances.Add(uniqueName, instance);

			return instance;
		}

		public static BbqpInstance GetInstance(string uniqueName, string shortName, Func<int[][]> produceQ, bool convertTo01 = false)
		{
			if (Instances.ContainsKey(uniqueName))
				return Instances[uniqueName];

			var filename = GetFullFilename(uniqueName);
			if (File.Exists(filename))
				return LoadFromFile(uniqueName, shortName, filename);

			if (!BbqpHelper.GenerationOfInstancesAllowed)
				throw new LabException("Generation of instances is prohibited.  Check the BbqpHelper.GenerationOfInstancesAllowed constant.  Attempted to generate " + filename);

			var timeCounter = AccurateRealTimeCounter.StartNew();
			var q = produceQ();
			var instance = new BbqpInstance(uniqueName,
			                               shortName,
			                               q,
			                               new long[q.Length],
			                               new long[q[0].Length]);
			timeCounter.Stop();
			//if (timeCounter.Elapsed > TimeInterval.Second)
				//Helper.DebugMessage("{0} spent on generating {1}.", timeCounter.Elapsed, shortName);
			//instance.SaveToFile(filename);

			if (convertTo01)
				instance = instance.ConvertTo01();

			Instances.Add(uniqueName, instance);

			return instance;
		}

		public int[] GetQRow(int i)
		{
			return q[i];
		}

		public long[] GetDArray()
		{
			return d;
		}

		public long[] GetCArray()
		{
			return c;
		}
	}
}
