﻿using System.Diagnostics;
using System.IO;

namespace BbqpCmcs.Types
{
	class BbqpSolution : Solution<BbqpInstance>
	{
		private readonly bool optimal;
		private readonly TimeInterval runningTime;
		private readonly bool[] x, y;
		private long objective = long.MinValue;

/*
		public BbqpSolution(BbqpInstance instance) : base(instance)
		{
			x = new bool[instance.N];
			y = new bool[instance.M];
		}
*/

		public BbqpSolution(BbqpInstance instance, string filename) : base(instance)
		{
			using (var reader = File.OpenText(filename))
			{
				var type = reader.ReadLine();
				switch (type.ToLower())
				{
					case "optimal":
						optimal = true;
						break;

					case "feasible":
						optimal = false;
						break;

					default:
						throw new LabException("Unknown solution type: {0}.", type);
				}

				runningTime = new TimeInterval(long.Parse(reader.ReadLine()));
				x = ReadBoolArray(reader.ReadLine());
				if (x.Length != instance.M)
					throw new LabException("Wrong solution dimension.");

				y = ReadBoolArray(reader.ReadLine());
				if (y.Length != instance.N)
					throw new LabException("Wrong solution dimension.");
			}
		}

		private bool[] ReadBoolArray(string line)
		{
			line = line.Trim();
			var result = new bool[line.Length];
			for (int i = 0; i < line.Length; i++)
				switch (line[i])
				{
					case '0':
						result[i] = false;
						break;

					case '1':
						result[i] = true;
						break;

					default:
						throw new LabException("Unknown bool value: {0}.", line[i]);
				}

			return result;
		}

		public BbqpSolution(BbqpInstance instance, bool[] x, bool[] y, TimeInterval runningTime, bool optimal) : base(instance)
		{
			this.x = x;
			this.y = y;
			this.runningTime = runningTime;
			this.optimal = optimal;

			Debug.Assert(x.Length == instance.M);
			Debug.Assert(y.Length == instance.N);
		}

		public BbqpSolution(BbqpInstance instance, bool[] x, bool[] y, TimeInterval runningTime, bool optimal, long objective) : base(instance)
		{
			this.x = x;
			this.y = y;
			this.runningTime = runningTime;
			this.optimal = optimal;
			this.objective = objective;

			Debug.Assert(x.Length == instance.M);
			Debug.Assert(y.Length == instance.N);
			Debug.Assert(CalculateObjective() == objective);
		}

		public BbqpSolution(BbqpSolution solution, TimeInterval runningTime)
			: this(solution.Instance, solution.x, solution.y, runningTime, solution.Optimal, solution.objective)
		{
		}

		public override void SaveToFile(string filename)
		{
			using (var writer = File.CreateText(filename))
			{
				writer.WriteLine(optimal ? "optimal" : "feasible");
				writer.WriteLine(runningTime.Usec);
				WriteArray(writer, x);
				WriteArray(writer, y);
			}
		}

		public void SaveToFileForTheLibrary(string filename)
		{
			using (var writer = File.CreateText(filename))
			{
				writer.WriteLine(optimal ? "optimal" : "feasible");
				writer.WriteLine(Objective);
				WriteArray(writer, x);
				WriteArray(writer, y);
			}
		}

		private void WriteArray(StreamWriter writer, bool[] values)
		{
			for (int i = 0; i < values.Length; i++)
				writer.Write(values[i] ? 1 : 0);
			writer.WriteLine();
		}

		public bool GetX(int index)
		{
			return x[index];
		}

/*
		public void SetX(int index, bool value)
		{
			x[index] = value;
		}
*/

		public bool GetY(int index)
		{
			return y[index];
		}

/*
		public void SetY(int index, bool value)
		{
			y[index] = value;
		}
*/

		public long Objective
		{
			get
			{
				if (objective == long.MinValue)
					objective = CalculateObjective();

				return objective;
			}
		}

		public long CalculateObjective()
		{
			long result = 0;

			for (int i = 0; i < instance.M; i++)
			{
				if (!x[i])
					continue;

				result += instance.GetC(i);

				for (int j = 0; j < instance.N; j++)
					if (y[j])
						result += instance[i, j];
			}

			for (int j = 0; j < instance.N; j++)
				if (y[j])
					result += instance.GetD(j);

			return result;
		}

		public TimeInterval RunningTime
		{
			get { return runningTime; }
		}

		public BbqpInstance Instance
		{
			get { return instance; }
		}

		public void CopyValues(out bool[] xCopy, out bool[] yCopy)
		{
			xCopy = (bool[])x.Clone();
			yCopy = (bool[])y.Clone();
		}

		public bool[] CopyX()
		{
			return (bool[])x.Clone();
		}

		public void CopyX(bool[] xCopy)
		{
			Debug.Assert(xCopy.Length == x.Length);
			x.CopyTo(xCopy, 0);
		}

		public void CopyY(bool[] yCopy)
		{
			Debug.Assert(yCopy.Length == y.Length);
			y.CopyTo(yCopy, 0);
		}

		public bool[] CopyY()
		{
			return (bool[])y.Clone();
		}


		public static BbqpSolution Trivial(BbqpInstance instance)
		{
			return new BbqpSolution(instance, new bool[instance.M], new bool[instance.N], TimeInterval.Zero, false);
		}

		public bool Optimal
		{
			get { return optimal; }
		}
	}
}
