﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net.Mail;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;

namespace BbqpCmcs.Types
{
	public static class Helper
	{
		public static bool TryParseTime(string time, out int result)
		{
			time = time.Trim();
			int pos = time.IndexOf(':');
			if (pos < 0)
			{
				result = 0;
				return false;
			}

			uint hours, minutes;
			if (!uint.TryParse(time.Substring(0, pos), out hours)
				|| !uint.TryParse(time.Substring(pos + 1), out minutes)
				|| minutes >= 60)
			{
				result = 0;
				return false;
			}

			result = (int)(hours * 60 + minutes);
			return true;
		}

		public static int ParseTime(string time)
		{
			time = time.Trim();
			int pos = time.IndexOf(':');
			if (pos < 0)
				throw new LabException();

			return Int32.Parse(time.Substring(0, pos)) * 60 + Int32.Parse(time.Substring(pos + 1));

			Match match = Regex.Match(time, @"\s*(\d+)\s*:\s*(\d\d)\s*");
			if (!match.Success)
				throw new LabException("Cannot parse time string '{0}'.", time);

			int h = Int32.Parse(match.Groups[1].Value);
			int m = Int32.Parse(match.Groups[2].Value);
			//if (h < 0 || h > 23)
			//	throw new LabException("Wrong hour in '{0}'.", time);
			if (m < 0 || m > 59)
				throw new LabException("Wrong minute in '{0}'.", time);
			return h * 60 + m;
		}

		public static string TimeToStr(int time)
		{
			if (time < 0)
				return "-" + TimeToStr(time == int.MinValue ? int.MaxValue : -time);

			int h = time / 60;
			int m = time - h * 60;
			return String.Format("{0:D2}:{1:00}", h, m);
		}

		public static string TimeSpanToStr(int time)
		{
			if (time < 0)
				return "-" + TimeSpanToStr(time == int.MinValue ? int.MaxValue : -time);

			int h = time / 60;
			int m = time - h * 60;
			return String.Format("{0}:{1:00}", h, m);
		}

		public static DateTime FromMinutes(int minutes)
		{
			return new DateTime(minutes * 600000000L);
		}

		public static void RandomPermutation<T>(T[] array, Random random)
		{
			for (int i = 0; i < array.Length - 2; i++)
			{
				int index = random.Next(i, array.Length);
				T temp = array[index];
				array[index] = array[i];
				array[i] = temp;
			}
		}

		public static void RandomPermutation<T>(List<T> list, Random random)
		{
			for (int i = 0; i < list.Count - 2; i++)
			{
				int index = random.Next(i, list.Count);
				T temp = list[index];
				list[index] = list[i];
				list[i] = temp;
			}
		}

		public static string ReadLine(TextReader reader, bool skipEmpty)
		{
			string line;
			do
			{
				line = reader.ReadLine();
				if (line == null)
					throw new LabException("Unexpected end of file.");

			} while (skipEmpty && String.IsNullOrWhiteSpace(line));

			return line;
		}

		public static readonly FileInfo ProgramFile = new FileInfo(Assembly.GetExecutingAssembly().Location);
		public static readonly DirectoryInfo ProgramDirectory = ProgramFile.Directory;


		public const int Delta = 5;

		public const int MinStayingTime = 5;

		public const string DefaultStyleFile = @"style.txt";

		public const int DefaultInPortTime = 10;

		public static int RequiredTransferTime = 10;
		public const int MaxTransferTime = 180;

		/*
				public static void Compile(Action<TextWriter> contentGenerator, OutputProperties outputProperties)
				{



					var viewer = new Process
					{
						StartInfo =
						{
							Arguments =
								"-1 \"" + ProgramDirectory.FullName + "\\" + filename + ".dvi\"",
							FileName =
								@"yap.exe"
						}
					};

					viewer.Start();
				}
		*/



		public static string ListToString(this IEnumerable list, string separator)
		{
			IEnumerator enumerator = list.GetEnumerator();
			if (!enumerator.MoveNext())
				return "";

			var builder = new StringBuilder(enumerator.Current.ToString());
			while (enumerator.MoveNext())
			{
				builder.Append(separator);
				builder.Append(enumerator.Current);
			}

			return builder.ToString();
		}

		public static string ListToString<T>(this IEnumerable<T> list, Func<T, object> selector, string separator)
		{
			IEnumerator<T> enumerator = list.GetEnumerator();
			if (!enumerator.MoveNext())
				return "";

			var builder = new StringBuilder(selector(enumerator.Current).ToString());
			while (enumerator.MoveNext())
			{
				builder.Append(separator);
				builder.Append(selector(enumerator.Current));
			}

			return builder.ToString();
		}

		public static bool ListEquals<T>(this List<T> list1, List<T> list2)
		{
			if (list1.Count != list2.Count)
				return false;

			for (int i = list1.Count - 1; i >= 0; i--)
			{
				if (!list1[i].Equals(list2[i]))
					return false;
			}

			return true;
		}

		public static bool ArrayEquals<T>(this T[] list1, T[] list2)
		{
			if (list1.Length != list2.Length)
				return false;

			for (int i = list1.Length - 1; i >= 0; i--)
			{
				if (!list1[i].Equals(list2[i]))
					return false;
			}

			return true;
		}

		public static int Random(Random random, int lessThan, int except)
		{
			int result = random.Next(lessThan - 1);

			result = result == except
				? lessThan - 1
				: result;

			Debug.Assert(result != except);
			Debug.Assert(result >= 0);
			Debug.Assert(result < lessThan);
			return result;
		}

		public static int Except(this Random random, int lessThan, int except)
		{
			Debug.Assert(lessThan > 1);

			int result = random.Next(lessThan - 1);

			result = result == except
				? lessThan - 1
				: result;

			Debug.Assert(result != except);
			Debug.Assert(result >= 0);
			Debug.Assert(result < lessThan);
			return result;
		}

		public const int StartOfTheDay = 0;
		public const int EndOfTheDay = 36 * 60;

		public static int Time(int hour, int minute)
		{
			return hour * 60 + minute;
		}

		public static void ErrorMessage(string format, params object[] args)
		{
			Message(ConsoleColor.Red, format, args);
		}

		public static void ErrorMessage(string message)
		{
			Message(ConsoleColor.Red, message);
		}

		public static int[] IntArray(int length, int value)
		{
			var result = new int[length];
			for (int i = length - 1; i >= 0; i--)
				result[i] = value;

			return result;
		}

		public static int Random(Random random, int lessThan, int except1, int except2)
		{
			if (except1 == except2)
				return Random(random, lessThan, except1);

			if (except1 == lessThan - 1)
				return Random(random, lessThan - 1, except2);

			int result1 = Random(random, lessThan - 1, except1);
			int result2 = result1 == except2
				? lessThan - 1
				: result1;

			Debug.Assert(result2 != except1);
			Debug.Assert(result2 != except2);
			Debug.Assert(result2 >= 0);
			Debug.Assert(result2 < lessThan);
			return result2;
		}

		public static int Except(this Random random, int lessThan, int except1, int except2)
		{
			return Random(random, lessThan, except1, except2);
		}

		[Obsolete]
		public static Random CreateRandom()
		{
			return new Random(123);
		}

		public static T Random<T>(this IEnumerable<T> list, Random random)
		{
			Debug.Assert(list.Any());
			var count = list.Count();
			int index = random.Next(count);
			return list.ElementAt(index);
		}

		public static T Random<T>(this List<T> list, Random random)
		{
			Debug.Assert(list.Count > 0);
			var count = list.Count;
			int index = random.Next(count);
			return list[index];
		}

		public static T Random<T>(this T[] array, Random random)
		{
			Debug.Assert(array.Length > 0);
			int index = random.Next(array.Length);
			return array[index];
		}

		public static int ParseTimeOrMinutes(string str)
		{
			if (str.IndexOf(':') < 0)
				return Int32.Parse(str);

			return ParseTime(str);
		}

		public static void Message(string message)
		{
			Message(ConsoleColor.White, message);
		}

		public static void Message(string message, params object[] args)
		{
			Message(ConsoleColor.White, String.Format(message, args));
		}

		private static object lockObject = new object();

		public static void Message(ConsoleColor color, string message)
		{
			lock (lockObject)
			{
				var old = Console.ForegroundColor;
				Console.ForegroundColor = color;
				Console.WriteLine(message);
				Console.ForegroundColor = old;
			}
		}

		public static void Message(ConsoleColor color, string format, params object[] args)
		{
			Message(color, String.Format(format, args));
		}

		public static void InfoMessage(string message)
		{
			Message(ConsoleColor.Yellow, message);
		}

		public static void InfoMessage(string format, params object[] args)
		{
			Message(ConsoleColor.Yellow, format, args);
		}

		public static void DebugMessage(string message)
		{
			Message(ConsoleColor.Green, message);
		}

		public static void DebugMessage(string format, params object[] args)
		{
			Message(ConsoleColor.Green, format, args);
		}

		public static string Underline(string value, bool condition)
		{
			return condition
				? "\\underline{" + value + "}"
				: value;
		}

		public static void Swap<T>(ref T x, ref T y)
		{
			T temp = x;
			x = y;
			y = temp;
		}

		private static CultureInfo culture;

		static Helper()
		{
			culture = CultureInfo.CreateSpecificCulture("en-GB");
			culture.DateTimeFormat.ShortDatePattern = "dd/MM/yyyy";
			Console.OutputEncoding = Encoding.UTF8;

			try
			{
				Console.SetBufferSize(210, 10000);
			}
			catch
			{
				return;
			}

			for (int height = 70; height >= 30; height--)
			{
				try
				{
					Console.SetWindowSize(210, height);
					break;
				}
				catch
				{
				}
			}
		}

		public static DateTime ParseDateTime(string str)
		{
			try
			{
				return DateTime.Parse(str, culture, DateTimeStyles.AssumeLocal);
			}
			catch
			{
				throw new LabException("Wrong date/time format: '{0}'.", str);
			}
		}

		public static T Max<T>(T x, T y) where T : IComparable<T>
		{
			return x.CompareTo(y) > 0 ? x : y;
		}

		public static T Min<T>(T x, T y) where T : IComparable<T>
		{
			return x.CompareTo(y) < 0 ? x : y;
		}

		public static DateTime Random(Random random, DateTime min, DateTime max)
		{
			return min + new TimeSpan((long)(random.NextDouble() * (max - min).Ticks));
		}

		public static bool ListsAreEqual<T>(List<T> list1, List<T> list2)
		{
			if (list1.Count != list2.Count)
				return false;

			for (int i = 0; i < list1.Count; i++)
				if (!list1[i].Equals(list2[i]))
					return false;

			return true;
		}

		public static int Sqr(int x)
		{
			return x * x;
		}

		public static float Sqr(float x)
		{
			return x * x;
		}

		public static double Sqr(double x)
		{
			return x * x;
		}

		public static DirectoryInfo GetSubdirectory(DirectoryInfo parent, params string[] names)
		{
			DirectoryInfo directory = parent;
			for (int i = 0; i < names.Length; i++)
				directory = GetSubdirectory(directory, names[i]);

			return directory;
		}

        public static DirectoryInfo GetSubdirectory(string parent, params string[] names)
		{
            var parentDir = new DirectoryInfo(parent);
            if (!parentDir.Exists)
                throw new LabException("Directory {0} does not exist.", parent);

            return GetSubdirectory(parentDir, names);
		}

	    public static DirectoryInfo GetSubdirectory(string parent, string name)
	    {
	        var parentDir = new DirectoryInfo(parent);
            if (!parentDir.Exists)
                throw new LabException("Directory {0} does not exist.", parent);

	        return GetSubdirectory(parentDir, name);
	    }

	    public static DirectoryInfo GetSubdirectory(DirectoryInfo parent, string name)
		{
			DirectoryInfo[] directories = parent.GetDirectories(name);
			switch (directories.Length)
			{
				case 0:
					return parent.CreateSubdirectory(name);

				case 1:
					return directories[0];

				default:
					throw new LabException("Several directories with the same name found.");
			}
		}

		public static string GetFullFilename(DirectoryInfo directory, string filename)
		{
			return directory.FullName + '/' + filename;
		}

		public static FileInfo GetFileInfo(DirectoryInfo directory, string filename)
		{
			return new FileInfo(GetFullFilename(directory, filename));
		}

		public static string GetFullFilename(string subdirName, string filename)
		{
			return GetFullFilename(GetSubdirectory(ProgramDirectory, subdirName), filename);
		}

		public static string GetFullFilename(string filename)
		{
			return GetFullFilename(ProgramDirectory, filename);
		}

		public static FileInfo CreateFile(DirectoryInfo directory, string fileName)
		{
			return new FileInfo(GetFullFilename(directory, fileName));
		}

		private struct Item<T> : IComparable<Item<T>> where T : IComparable
		{
			private readonly int index;
			private readonly T element;

			public Item(int index, T element)
			{
				this.index = index;
				this.element = element;
			}

			public int CompareTo(Item<T> other)
			{
				return element.CompareTo(other.element);
			}

			public int Index
			{
				get { return index; }
			}
		}

		public static int[] GetSortedOrder<T>(T[] array, bool descending = false) where T : IComparable
		{
			var items = new Item<T>[array.Length];
			for (int i = 0; i < array.Length; i++)
				items[i] = new Item<T>(i, array[i]);

			Array.Sort(items);

			var result = new int[array.Length];
			if (@descending)
				for (int i = 0; i < array.Length; i++)
					result[i] = items[array.Length - 1 - i].Index;
			else
				for (int i = 0; i < array.Length; i++)
					result[i] = items[i].Index;

			return result;
		}

		public static double NormallyDistributedRandom(Random random)
		{
			var u = random.NextDouble();
			var v = random.NextDouble();
			return Math.Sqrt(-2 * Math.Log(u)) * Math.Cos(2 * Math.PI * v);
		}

		public static void PerformAsync(CountdownEvent countdown, Action action)
		{
			countdown.AddCount();
			ThreadPool.QueueUserWorkItem(
				state =>
				{
					try
					{
						action();
					}
					finally
					{
						countdown.Signal();
					}
				});
		}

		public static readonly char[] Tab = new[] { '\t' };
		public static readonly char[] SpaceAndTab = new[] { ' ', '\t' };

		public static string[] ReadAndSplit(TextReader reader, int count)
		{
			var line = ReadLine(reader, false);

			var strings = line.Split(SpaceAndTab, StringSplitOptions.RemoveEmptyEntries);
			if (strings.Length != count)
				throw new LabException(
					"{0} entries were expected but {1} are found in line '{1}'.",
					count,
					strings.Length,
					line);

			return strings;
		}

		public static string GetFullFilename(DirectoryInfo directory, string filename, string extension)
		{
			return GetFullFilename(directory, filename + extension);
		}

		public static void WriteTabSeparatedLine(TextWriter writer, IEnumerable values)
		{
			var enumerator = values.GetEnumerator();
			if (enumerator.MoveNext())
			{
				writer.Write("{0}", enumerator.Current);
				while (enumerator.MoveNext())
					writer.Write("\t{0}", enumerator.Current);
			}

			writer.WriteLine();
		}

		public static void WriteTabSeparatedLine(TextWriter writer, params object[] values)
		{
			for (int i = 0; i < values.Length - 1; i++)
				writer.Write("{0}\t", values[i]);

			writer.WriteLine(values[values.Length - 1]);
		}

		public static T[] GetEnumElements<T>()
		{
			return (T[])Enum.GetValues(typeof(T));
		}

		public static int GetEnumElementsCount<T>()
		{
			return Enum.GetValues(typeof(T)).Length;
		}

		public static string NameNoExt(this FileInfo fileInfo)
		{
			return fileInfo.Name.Substring(0, fileInfo.Name.Length - fileInfo.Extension.Length);
		}

		public static int[] DistinctUnordered(this Random random, int lessThan, int numberOfElements)
		{
			Debug.Assert(numberOfElements <= lessThan);
			switch (numberOfElements)
			{
				case 0:
					return new int[0];

				case 1:
					return new[] { random.Next(lessThan) };

				case 2:
				{
					var result = new int[2];
					result[0] = random.Next(lessThan);
					result[1] = random.Except(lessThan, result[0]);
					return result;
				}

				case 3:
				{
					var result = new int[3];
					result[0] = random.Next(lessThan);
					result[1] = random.Except(lessThan, result[0]);
					result[2] = random.Except(lessThan, result[0], result[1]);
					return result;
				}

				default:
				{
					if (2 * numberOfElements >= lessThan)
					{ // O(numberOfElements * lessThan) iterations; O(numberOfElements + lessThan) memory
						var available = new Range(lessThan).ToList();
						var result = new int[numberOfElements];
						for (int i = 0; i < numberOfElements; i++)
						{
							int index = random.Next(available.Count);
							result[i] = available[index];
							available.RemoveAt(index);
						}

						return result;
					}
					else
					{ // O(numberOfElements^2) iterations if (numberOfElements / lessThan < const < 1); O(numberOfElements) memory
						var result = new int[numberOfElements];
						for (int i = 0; i < numberOfElements; i++)
						{
							int value;
							do
							{
								value = random.Next(lessThan);
							} while (Array.IndexOf(result, value, 0, i) >= 0);

							result[i] = value;
						}

						return result.ToArray();
					}
				}
			}
		}

/*
		public static T[] NewArray<T>(int length, T fillWith)
		{
			var result = new T[length];
			for (int i = result.Length - 1; i > 0; i--)
				result[i] = fillWith;

			return result;
		}
*/

		public static T[] FillArray<T>(this T fillWith, int length)
		{
			var result = new T[length];
			for (int i = result.Length - 1; i >= 0; i--)
				result[i] = fillWith;

			return result;
		}

		public static void FillExistingArray<T>(T[] array, T fillWith)
		{
			for (int i = array.Length - 1; i >= 0; i--)
				array[i] = fillWith;
		}

		public static void RunConcurrent(int threadCount, Action<int> threadProc)
		{
			int counter = threadCount;
			for (int i = 0; i < threadCount; i++)
			{
				int threadIndex = i;
				new Thread(
					() =>
					{
						threadProc(threadIndex);
						Interlocked.Decrement(ref counter);
					}).Start();
			}

			while (counter > 0)
				Thread.Sleep(10);
		}

		public static bool NextBoolean(this Random random)
		{
			return random.Next(2) == 0;
		}

		public static bool NextBoolean(this Random random, double trueProbability)
		{
			return random.NextDouble() < trueProbability;
		}

		public static bool[] BitMaskToBoolArray(this int value, int maskLength)
		{
			var result = new bool[maskLength];
			for (int i = 0; i < result.Length; i++)
				result[i] = (maskLength << i) != 0;

			return result;
		}

		public static string Repeat(this string s, int numberOfTimes)
		{
			Debug.Assert(numberOfTimes >= 0);

			switch (numberOfTimes)
			{
				case 0:
					return "";

				case 1:
					return s;

				default:
					var stringBuilder = new StringBuilder(s.Length * numberOfTimes);
					for (int i = 0; i < numberOfTimes; i++)
						stringBuilder.Append(s);

					return stringBuilder.ToString();
			}
		}

		public static IEnumerable<DayOfWeek> DaysOfWeek(DayOfWeek first = DayOfWeek.Monday)
		{
			for (int i = 0; i < 7; i++)
				yield return (DayOfWeek)((i + (int)first) % 7);
		}

		public static T Roulette<T>(this Random random, WeightedEntity<T>[] entities)
		{
			int totalWeight = 0;
			for (int j = 0; j < entities.Length; j++)
				totalWeight += entities[j].Weight;

			var r = random.Next(totalWeight);
			int accumulated = 0;
			int i = 0;
			do
			{
				accumulated += entities[i++].Weight;
			} while (accumulated <= r);

			return entities[i - 1].Entity;
		}

		public static int Roulette(this Random random, double[] weights)
		{
#if DEBUG
		    try
#endif
		    {
			    double totalWeight = 0;
		        for (int j = 0; j < weights.Length; j++)
		            totalWeight += weights[j];

		        var r = random.NextDouble()*totalWeight;
		        double accumulated = 0;
		        int i = 0;
		        do
		        {
		            accumulated += weights[i++];
		        } while (accumulated <= r);
            
                Debug.Assert(i <= weights.Length);
                return i - 1;
            }
#if DEBUG
		    catch
		    {
                Helper.ErrorMessage("Roulette exception.  {0}", weights.ListToString(", "));
                return 0;
		    }
#endif
		}

		public static int RouletteProbabilities(this Random random, double[] probabilities)
		{
			Debug.Assert(probabilities.Length > 0 && Math.Abs(probabilities.Sum() - 1) < 1e-6 && probabilities.All(w => w >= 0));

			var r = random.NextDouble();
			double accumulated = 0;

			int i;
			for (i = probabilities.Length - 1; i > 0; i--)
			{
				accumulated += probabilities[i];
				if (accumulated > r)
					break;
			}
			
			return i;
		}

		public static int RouletteProbabilities(this Random random, double[] probabilities, int firstIndex, int numberOfElements)
		{
			Debug.Assert(probabilities.Length > 0 && Math.Abs(new Range(firstIndex, firstIndex + numberOfElements - 1).Sum(j => probabilities[j])) < 1e-6 && probabilities.All(w => w >= 0));

			var r = random.NextDouble();
			double accumulated = 0;

			int i;
			for (i = firstIndex + numberOfElements - 1; i > firstIndex; i--)
			{
				accumulated += probabilities[i];
				if (accumulated > r)
					break;
			}
			
			return i;
		}

		public static int RouletteProbabilities(this Random random, float[] probabilities, int firstIndex, int numberOfElements)
		{
			Debug.Assert(probabilities.Length > 0 && Math.Abs(new Range(firstIndex, firstIndex + numberOfElements - 1).Sum(j => probabilities[j])) < 1e-6 && probabilities.All(w => w >= 0));

			var r = random.NextDouble();
			double accumulated = 0;

			int i;
			for (i = firstIndex + numberOfElements - 1; i > firstIndex; i--)
			{
				accumulated += probabilities[i];
				if (accumulated > r)
					break;
			}
			
			return i;
		}

		private static readonly Dictionary<string, DateTime> lastTimePrinted = new Dictionary<string, DateTime>();

		public static bool EnsureDebugInfoInterval(string label, TimeInterval minInterval)
		{
			var now = DateTime.Now;
			if (!lastTimePrinted.ContainsKey(label))
				lock (lastTimePrinted)
				{
					// By the time the thread gets the lock, some other thread could add the label
					if (!lastTimePrinted.ContainsKey(label))
						lastTimePrinted.Add(label, now);
				}
			else
			{
				var lastPrinted = lastTimePrinted[label];
				if (now - lastPrinted >= (TimeSpan)minInterval)
				{
					lastTimePrinted[label] = now;
					return true;
				}
			}

			return false;
		}

/*
		public static void DebugInfoWithMinInterval(
			TimeInterval minInterval,
			string label,
			string format,
			params object[] args)
		{
			var now = DateTime.Now;
			if (!lastTimePrinted.ContainsKey(label))
				lastTimePrinted.Add(label, now);
			else
			{
				var lastPrinted = lastTimePrinted[label];
				if (now - lastPrinted >= (TimeSpan)minInterval)
				{
					lastTimePrinted[label] = now;
					DebugMessage(format, args);
				}
			}
		}
*/

		public static readonly bool[] FalseTrue = { false, true };
		public static readonly bool[] TrueFalse = { true, false };


		private static List<int> factorials = new List<int>(new[] { 1 });

		static int Factorial(int n)
		{
			if (n < factorials.Count)
				return factorials[n];

			int result;
			lock (factorials)
			{
				result = Factorial(n - 1) * n;

				Debug.Assert(factorials.Count == n);
				factorials.Add(result);
			}

			return result;
		}

		public static int Choose(int n, int k)
		{
			switch (k)
			{
				case 1:
					return n;

				case 2:
					return n * (n - 1) / 2;

				case 3:
					return n * (n - 1) * (n - 2) / 6;

				default:
					return Factorial(n) / (Factorial(k) * Factorial(n - k));
			}
		}


		public static T[][] ToArrayOfArrays<T>(T[,] array)
		{
			var rows = array.GetLength(0);
			var cols = array.GetLength(1);

			var result = new T[rows][];
			for (int i = result.Length - 1; i >= 0; i--)
			{
				result[i] = new T[cols];
				for (int j = result[i].Length - 1; j >= 0; j--)
					result[i][j] = array[i, j];
			}

			return result;
		}

		public static T[,] To2DArray<T>(T[][] array)
		{
			if (array.Length == 0)
				return new T[0, 0];

			int rows = array.Length;
			int cols = array[0].Length;

			var result = new T[rows, cols];
			for (int i = array.Length - 1; i >= 0; i--)
			{
				Debug.Assert(array[i].Length == cols);
				for (int j = array[i].Length - 1; j >= 0; j--)
					result[i, j] = array[i][j];
			}

			return result;
		}



		private static int numberOfPhysicalCores = 0;

		public static bool Odd(int value)
		{
			return (value & 1) != 0;
		}

		public static void ForEach<T>(this IEnumerable<T> list, Action<T> action)
		{
			foreach (var element in list)
				action(element);
		}

		public static double MedianUnsorted(this double[] values)
		{
			var clone = (double[])values.Clone();
            Array.Sort(clone);
			return MedianSorted(clone);
		}

		public static double MedianSorted(this double[] values)
		{
			if (values.Length == 0)
				throw new LabException("The median cannot be calculated for an empty array.");

			if (Odd(values.Length))
				return values[values.Length / 2];

			return (values[values.Length / 2 - 1] + values[values.Length / 2]) / 2;
		}

		public static double Median(this IEnumerable<double> values)
		{
			var array = values.ToArray();
			Array.Sort(array);
			return MedianSorted(array);
		}

		public static double Percentile(this IEnumerable<double> values, double percentile)
		{
			var array = values.ToArray();
			Array.Sort(array);
			return PercentileSorted(array, percentile);
		}

		public static double PercentileUnsorted(this double[] values, double percentile)
		{
			var clone = (double[])values.Clone();
			Array.Sort(clone);
			return PercentileSorted(clone, percentile);
		}

		public static double PercentileSorted(this double[] values, double percentile)
		{
			Debug.Assert(percentile >= 0 && percentile <= 1);
			if (values.Length == 0)
				throw new LabException("The percentile cannot be calculated for an empty array.");

			var realIndex = (values.Length - 1) * percentile;
			int indexBelow = (int)realIndex;
			int indexAbove = indexBelow + 1;
			if (indexAbove >= values.Length)
				return values[indexBelow];

			double valueBelow = values[indexBelow];
			double valueAbove = values[indexAbove];
			double deltaY = valueAbove - valueBelow;
			double x = realIndex - indexBelow;
			
			return valueBelow + deltaY * x;
		}

		public static void Split<T>(
			this IEnumerable<T> original,
			Predicate<T> condition,
			List<T> falseList,
			List<T> trueList)
		{
			foreach (var value in original)
			{
				if (condition(value))
					trueList.Add(value);
				else
					falseList.Add(value);
			}
		}

		public static void Split<T>(
			this IEnumerable<T> original,
			Predicate<T> condition,
			out List<T> falseList,
			out List<T> trueList)
		{
			falseList = new List<T>();
			trueList = new List<T>();
			original.Split(condition, falseList, trueList);
		}

		public static void SendNotification(string subject, string body)
		{
		    try
		    {
		        var client = new SmtpClient
		        {
		            Port = 587,
		            Host = "smtp.gmail.com",
		            EnableSsl = true,
		            Timeout = 1000,
		            DeliveryMethod = SmtpDeliveryMethod.Network,
		            UseDefaultCredentials = false,
		            Credentials = new System.Net.NetworkCredential("karapetyan.robot@gmail.com", "IAmAMachine")
		        };

		        var mm = new MailMessage("karapetyan.robot@gmail.com", "daniel.karapetyan@gmail.com", subject, body)
		        {
		            BodyEncoding = Encoding.UTF8,
		            DeliveryNotificationOptions = DeliveryNotificationOptions.OnFailure
		        };

		        client.Send(mm);
		    }
		    catch (Exception ex)
		    {
		        ErrorMessage("Failed to send an email: {0}", ex);
		    }
		}

		public static void WriteCsvLine(StreamWriter writer, params object[] values)
		{
			writer.WriteLine(values.ListToString(v => "\"" + v + '"', ","));
		}

		public static bool SubdirectoryExists(DirectoryInfo parent, string subdirectory)
		{
			switch (parent.GetDirectories(subdirectory).Length)
			{
				case 0:
					return false;

				case 1:
					return true;

				default:
					throw new LabException("Several matching subdirectories found.");
			}
		}

		public static string TextBar(double value, int width)
		{
            if (value < 0 || value > 1.0)
                throw new LabException("Out of bounds value for text bar.");

			value *= width;
			var result = ' '.FillArray(width);
			var integer = (int)value;
			for (int i = 0; i < integer; i++)
				result[i] = '█';

			value -= integer;
			if (value < 0.25)
			{
			}
			else if (value <= 0.75)
				result[integer] = '▌';
			else
				result[integer] = '█';

			return new string(result);
		}

	    public static void RemoveLast<T>(this List<T> list)
	    {
	        list.RemoveAt(list.Count - 1);
	    }
	}
}
